Package: functionals
Type: Package
Title: Functional Programming with Parallelism and Progress Tracking
Version: 0.5.0
Authors@R: 
    person("Imad", "EL BADISY", email = "elbadisyimad@gmail.com", role = c("aut", "cre"))
Maintainer: Imad EL BADISY <elbadisyimad@gmail.com>
Description: Provides functional tools such as fmap(), fwalk(), and fapply() 
    to iterate over vectors, data frames, or grouped data with optional parallelism 
    and real-time progress tracking. Designed for readable and reproducible workflows, 
    including support for Monte Carlo simulations and benchmarking.
License: MIT + file LICENSE
Encoding: UTF-8
Imports: parallel
Suggests: testthat (>= 3.0.0), bench, rsample
Config/testthat/edition: 3
RoxygenNote: 7.3.2
BugReports: https://github.com/ielbadisy/functionals/issues
NeedsCompilation: no
Packaged: 2025-07-15 12:26:32 UTC; imad-el-badisy
Author: Imad EL BADISY [aut, cre]
Repository: CRAN
Date/Publication: 2025-07-18 15:00:08 UTC
Built: R 4.6.0; ; 2026-04-26 18:36:43 UTC; unix

Package: functionals
Type: Package
Title: Functional Programming with Parallelism and Progress Tracking
Version: 0.5.1
Authors@R: 
    person("Imad", "El Badisy", email = "elbadisyimad@gmail.com", role = c("aut", "cre"))
Maintainer: Imad El Badisy <elbadisyimad@gmail.com>
Description: Provides functional tools such as fmap(), fwalk(), and fapply() 
    to iterate over vectors, data frames, or grouped data with optional parallelism 
    and real-time progress tracking. Progress updates now reflect completed tasks
    across sequential, multicore, and cluster-backed execution. Designed for
    readable and reproducible workflows, including support for Monte Carlo
    simulations and benchmarking.
License: MIT + file LICENSE
Encoding: UTF-8
Imports: parallel
Suggests: testthat (>= 3.0.0), bench, rsample
Config/testthat/edition: 3
RoxygenNote: 7.3.3
URL: https://CRAN.R-project.org/package=functionals
BugReports: https://github.com/ielbadisy/functionals/issues
NeedsCompilation: no
Packaged: 2026-08-21 09:10:07 UTC; imad-el-badisy
Author: Imad El Badisy [aut, cre]
Repository: CRAN
Date/Publication: 2026-08-22 02:10:02 UTC

# Generated by roxygen2: do not edit by hand

export(fapply)
export(fcompose)
export(fcv)
export(floop)
export(fmap)
export(fmapc)
export(fmapg)
export(fmapn)
export(fmapr)
export(freduce)
export(frepeat)
export(fwalk)
importFrom(utils,flush.console)

# functionals: Functional mapping with parallelism and progress bars

## Overview

`functionals` is a lightweight toolkit for functional programming in R
with built-in support for parallelism and progress bars. It extends base
R’s functional tools with a consistent, minimal API for mapping,
walking, reducing, cross-validating, and repeating computations across
lists, data frames, and grouped data.

As of `0.5.1`, progress reporting is completion-driven across
sequential, multicore, and cluster-backed execution. When `pb = TRUE`,
the bar advances as individual tasks finish rather than at internal
chunk boundaries. Rendering is throttled for large workloads so the
console keeps a single lightweight status bar instead of redrawing on
every task.

## Function Reference Table

| Function     | Main arguments                        | Output type | Description                                                 |
|--------------|---------------------------------------|-------------|-------------------------------------------------------------|
| `fmap()`     | `.x`, `.f`, `ncores`, `pb`            | list        | Map `.f` over elements of `.x`                              |
| `fmapn()`    | `.l`, `.f`, `ncores`, `pb`            | list        | Map `.f` over multiple aligned lists                        |
| `fmapr()`    | `.df`, `.f`, `ncores`, `pb`           | list        | Map `.f` over each row of a data frame (as named list)      |
| `fmapc()`    | `.df`, `.f`, `ncores`, `pb`           | list        | Map `.f(column, name)` over each column                     |
| `fmapg()`    | `.df`, `.f`, `by`, `ncores`, `pb`     | list        | Map `.f(group_df)` over groups defined by a column          |
| `floop()`    | `.x`, `.f`, `...`, `ncores`, `pb`     | list        | General-purpose functional loop with side-effects           |
| `fwalk()`    | `.x`, `.f`, `ncores`, `pb`            | NULL        | Map `.f` over `.x` for side-effects only (invisible return) |
| `frepeat()`  | `times`, `expr`, `.x`, `ncores`, `pb` | list/vector | Repeat a call/expression multiple times                     |
| `fcv()`      | `.splits`, `.f`, `ncores`, `pb`       | list        | Map `.f` over resampling splits from `rsample::vfold_cv()`  |
| `freduce()`  | `.x`, `.f`, `...`                     | scalar/list | Reduce `.x` using a binary function `.f`                    |
| `fcompose()` | any number of functions `f1, f2, ...` | function    | Compose multiple functions: `f1(f2(...(x)))`                |
| `fapply()`   | `.x`, `.f`, `ncores`, `pb`, `...`     | list        | Core internal utility for applying a function over `.x`     |

## Syntax Equivalence

| Task                     | `functionals` Example                                     | `purrr` Example                                    | Base R                                      |
|--------------------------|-----------------------------------------------------------|----------------------------------------------------|---------------------------------------------|
| Map square               | `fmap(1:5, function(x) x^2)`                              | `map(1:5, function(x) x^2)`                        | `lapply(1:5, function(x) x^2)`              |
| Map over N arguments     | `fmapn(list(1:3, 4:6, 7:9), function(x, y, z) x + y + z)` | `pmap(list(1:3, 4:6, 7:9), function(x, y, z) ...)` | `Map(function(x, y, z) ..., 1:3, 4:6, 7:9)` |
| Map over data frame rows | `fmapr(df, function(row) row$a + row$b)`                  | `pmap(df[c("a", "b")], function(x, y) x + y)`      | `apply(df, 1, function(row) ...)`           |
| Map over data frame cols | `fmapc(df, function(x, name) mean(x))`                    | `imap(df, function(x, name) mean(x))`              | `lapply(df, mean)`                          |
| Grouped map              | `fmapg(df, f, by = "group")`                              | `map(split(df, df$group), f)`                      | `lapply(split(df, df$group), f)`            |
| General-purpose loop     | `floop(1:3, function(x) cat(x))`                          | *(manual recursion)*                               | `for (x in 1:3) cat(x)`                     |
| Parallel + progress      | `fmap(x, f, ncores = 4, pb = TRUE)`                       | *(future_map(x, f))* with `progressr`              | `parLapply(cl, x, f)` or `mclapply()`       |
| Repeat simulation        | `frepeat(100, function() rnorm(1))`                       | *(manual loop)*                                    | `replicate(100, rnorm(1))`                  |
| Walk with side effects   | `fwalk(letters, function(x) cat(x))`                      | `walk(letters, function(x) cat(x))`                | `lapply(letters, cat)`                      |
| Reduce                   | `` freduce(1:5, `+`) ``                                   | `` reduce(1:5, `+`) ``                             | `` Reduce(`+`, 1:5) ``                      |
| Compose functions        | `fcompose(sqrt, abs)(-4)`                                 | `compose(sqrt, abs)(-4)`                           | `(function(x) sqrt(abs(x)))(-4)`            |

## Why no formula interface like `~ .x + .y`?

While `functionals` draws inspiration from `purrr`, it intentionally
avoids supporting the formula-based anonymous function syntax (e.g.,
`~ .x + 1`) for now.

This decision is based on:

- Keeping dependencies minimal (no reliance on `rlang`)
- Avoiding non-standard evaluation that can confuse new users
- Encouraging explicit, readable code using `function(x) { ... }` style

We may consider adding tidy evaluation support (e.g., with quosures or
`rlang::as_function`) in a future release. However, the current
philosophy favors clarity and simplicity.

## Installation

``` r
# from CRAN
install.packages("functionals")

# from github
remotes::install_github("ielbadisy/functionals")
```

## Examples

``` r
library(functionals)
library(purrr)
library(furrr)
#> Loading required package: future
library(pbapply)
library(dplyr)
#> 
#> Attaching package: 'dplyr'
#> The following objects are masked from 'package:stats':
#> 
#>     filter, lag
#> The following objects are masked from 'package:base':
#> 
#>     intersect, setdiff, setequal, union
library(rsample)
library(bench)


plan(multisession)

# utility to compare results
compare_outputs <- function(label, x, y) {
  cat("\n", label, "->", if (identical(x, y)) "dentical\n" else if (isTRUE(all.equal(x, y))) "nearly equal\n" else "different\n")
}

# strip names and convert to plain numeric vector
as_vec <- function(x) as.numeric(unlist(x, use.names = FALSE))
```

### Element-wise map

``` r
x1 <- fmap(1:5, function(x) x^2)
x2 <- lapply(1:5, function(x) x^2)
x3 <- map(1:5, ~ .x^2)
x4 <- future_map(1:5, ~ .x^2)
x5 <- pblapply(1:5, function(x) x^2)
compare_outputs("Element-wise: base", x1, x2)
#> 
#>  Element-wise: base -> dentical
compare_outputs("Element-wise: purrr", x1, x3)
#> 
#>  Element-wise: purrr -> dentical
compare_outputs("Element-wise: furrr", x1, x4)
#> 
#>  Element-wise: furrr -> dentical
compare_outputs("Element-wise: pbapply", x1, x5)
#> 
#>  Element-wise: pbapply -> dentical
```

### Parallel progress semantics

``` r
slow <- function(x) {
  Sys.sleep(c(0.12, 0.24, 0.36, 0.48)[[x]])
  x^2
}

fmap(1:4, slow, ncores = 2, pb = TRUE)
#>                                                                                 [                              ]   0% 0/4 elapsed 00:00                                                                                [=======                       ]  25% 1/4 elapsed 00:00 eta 00:00                                                                                [===============               ]  50% 2/4 elapsed 00:00 eta 00:00                                                                                [======================        ]  75% 3/4 elapsed 00:00 eta 00:00                                                                                [==============================] 100% 4/4 elapsed 00:01 eta 00:00
#> [[1]]
#> [1] 1
#> 
#> [[2]]
#> [1] 4
#> 
#> [[3]]
#> [1] 9
#> 
#> [[4]]
#> [1] 16
```

The progress bar advances on each completed task, shows elapsed time and
ETA, and throttles redraws for large workloads so intermediate counts
stay readable without repainting the console on every iteration.

## Multi-input map

``` r
x1 <- fmapn(list(1:3, 4:6), function(x, y) x + y)
x2 <- Map(`+`, 1:3, 4:6)
x3 <- pmap(list(1:3, 4:6), ~ ..1 + ..2)
x4 <- future_pmap(list(1:3, 4:6), ~ ..1 + ..2)
compare_outputs("Multi-input: base", x1, x2)
#> 
#>  Multi-input: base -> dentical
compare_outputs("Multi-input: purrr", x1, x3)
#> 
#>  Multi-input: purrr -> dentical
compare_outputs("Multi-input: furrr", x1, x4)
#> 
#>  Multi-input: furrr -> dentical
```

### Row-wise map

``` r
x1 <- fmapr(mtcars, function(row) row$mpg + row$cyl)
rowlist <- lapply(seq_len(nrow(mtcars)), function(i) as.list(mtcars[i, ]))
x2 <- lapply(rowlist, function(row) row$mpg + row$cyl)
x3 <- map(rowlist, function(row) row$mpg + row$cyl)
compare_outputs("Row-wise: base", as_vec(x1), as_vec(x2))
#> 
#>  Row-wise: base -> dentical
compare_outputs("Row-wise: purrr", as_vec(x1), as_vec(x3))
#> 
#>  Row-wise: purrr -> dentical
```

### Column-wise map

``` r
x1 <- fmapc(mtcars, function(col, name) mean(col))
x2 <- sapply(mtcars, mean)
x3 <- imap(mtcars, ~ mean(.x))
x4 <- future_imap(mtcars, ~ mean(.x))
compare_outputs("Column-wise: base", x1, as.list(x2))
#> 
#>  Column-wise: base -> dentical
compare_outputs("Column-wise: purrr", x1, x3)
#> 
#>  Column-wise: purrr -> dentical
compare_outputs("Column-wise: furrr", x1, x4)
#> 
#>  Column-wise: furrr -> dentical
```

### Group-wise map

``` r
x1 <- fmapg(iris, function(df) colMeans(df[1:4]), by = "Species")
x2 <- lapply(split(iris, iris$Species), function(df) colMeans(df[1:4]))
x3 <- map(split(iris, iris$Species), ~ colMeans(.x[1:4]))
x4 <- future_map(split(iris, iris$Species), ~ colMeans(.x[1:4]))
compare_outputs("Group-wise: base", x1, x2)
#> 
#>  Group-wise: base -> dentical
compare_outputs("Group-wise: purrr", x1, x3)
#> 
#>  Group-wise: purrr -> dentical
compare_outputs("Group-wise: furrr", x1, x4)
#> 
#>  Group-wise: furrr -> dentical
```

### Side-effect map

``` r
cat("\nSide-effects:\n")
#> 
#> Side-effects:
fwalk(1:3, print)
#> [1] 1
#> [1] 2
#> [1] 3
```

### General-purpose loop with return values

``` r
x1 <- floop(1:5, function(x) x^2, .capture = TRUE)
x2 <- lapply(1:5, function(x) x^2)
x3 <- {
  out <- list()
  for (i in 1:5) out[[i]] <- i^2
  out
}
compare_outputs("floop() vs lapply()", x1, x2)
#> 
#>  floop() vs lapply() -> dentical
compare_outputs("floop() vs for()", x1, x3)
#> 
#>  floop() vs for() -> dentical
```

### General-purpose loop (side-effect only)

``` r
cat("\nGeneral-purpose loop (side-effects):\n")
#> 
#> General-purpose loop (side-effects):
floop(1:3, function(x) cat("floop says:", x, "\n"), pb = TRUE, .capture = FALSE)
#>                                                                                 [                              ]   0% 0/3 elapsed 00:00                                                                                floop says: 1 
#>                                                                                 [==========                    ]  33% 1/3 elapsed 00:00 eta 00:00                                                                                floop says: 2 
#>                                                                                 [====================          ]  67% 2/3 elapsed 00:00 eta 00:00                                                                                floop says: 3 
#>                                                                                 [==============================] 100% 3/3 elapsed 00:00 eta 00:00
cat("for-loop equivalent:\n")
#> for-loop equivalent:
for (x in 1:3) cat("for says:", x, "\n")
#> for says: 1 
#> for says: 2 
#> for says: 3
```

### Cross-validation

``` r
splits <- vfold_cv(iris, v = 3)$splits
fit_model <- function(split) mean(analysis(split)$Sepal.Length)
x1 <- fcv(splits, fit_model)
x2 <- lapply(splits, fit_model)
compare_outputs("CV map: base", x1, x2)
#> 
#>  CV map: base -> dentical
```

### Repeat simulation

``` r
x1 <- frepeat(times = 10, expr = rnorm(1))
x2 <- as.list(replicate(10, rnorm(1)))
x3 <- as.list(pbreplicate(10, rnorm(1)))
cat("\nRepeat: Results not comparable (randomized output)\n")
#> 
#> Repeat: Results not comparable (randomized output)
```

### Reduce

``` r
x1 <- freduce(1:5, `+`)
x2 <- Reduce(`+`, 1:5)
x3 <- reduce(1:5, `+`)
compare_outputs("Reduce: base", x1, x2)
#> 
#>  Reduce: base -> dentical
compare_outputs("Reduce: purrr", x1, x3)
#> 
#>  Reduce: purrr -> dentical
```

### Compose

``` r
x1 <- fcompose(sqrt, abs)(-4)
x2 <- (function(x) sqrt(abs(x)))(-4)
x3 <- compose(sqrt, abs)(-4)
compare_outputs("Compose: base", x1, x2)
#> 
#>  Compose: base -> dentical
compare_outputs("Compose: purrr", x1, x3)
#> 
#>  Compose: purrr -> dentical
```

#' Apply a function to groups of a data frame in parallel
#'
#' Applies a function `.f` to each group of rows in a data frame `.df`, where grouping is defined by one or more variables in `by`.
#' Each group is passed as a data frame to `.f`. Supports parallelism and optional progress display.
#'
#' @param .df A data frame to group and apply the function over.
#' @param .f A function to apply to each group. The function should accept a data frame (a group).
#' @param by A character vector of column names in `.df` used for grouping.
#' @param ncores Integer. Number of cores to use for parallel processing. Default is `NULL` (sequential).
#' @param pb Logical. Whether to show a progress bar. Default is `FALSE`.
#' @param ... Additional arguments passed to `.f`.
#'
#' @return A list of results, one for each group defined by `by`.
#'
#' @examples
#' # Group-wise mean of Sepal.Length in iris dataset
#' fmapg(iris, function(df) mean(df$Sepal.Length), by = "Species")
#'
#' # Group-wise model fitting with progress and parallelism
#' \donttest{
#' fmapg(mtcars, function(df) lm(mpg ~ wt, data = df),
#'       by = "cyl", ncores = 2, pb = TRUE)
#' }
#'
#' @export


fmapg <- function(.df, .f, by, ncores = NULL, pb = FALSE, ...) {
  stopifnot(is.data.frame(.df), all(by %in% names(.df)))
  groups <- split(.df, .df[by], drop = TRUE)
  .f <- match.fun(.f)
  fapply(groups, .f, ncores = ncores, pb = pb, ...)
}

test_that("fmapg applies function to grouped data", {
  df <- data.frame(
    id = rep(1:3, each = 2),
    value = c(10, 20, 30, 40, 50, 60)
  )

  out <- fmapg(df, function(group) sum(group$value), by = "id")

  expect_length(out, 3)
  expect_true(all(sapply(out, is.numeric)))
})

test_that("fmapg supports parallel execution with progress", {
  skip_on_cran()
  df <- data.frame(
    g = rep(letters[1:4], each = 5),
    x = rnorm(20)
  )

  out <- fmapg(df, function(gr) mean(gr$x), by = "g", ncores = 2, pb = TRUE)

  expect_equal(length(out), 4)
  expect_true(all(sapply(out, is.numeric)))
})

test_that("fmapg errors with missing group variables", {
  df <- data.frame(a = 1:3)
  expect_error(fmapg(df, identity, by = "missing_column"))
})

#' Apply a function row-wise on a data frame with parallelism
#'
#' Applies a function `.f` to each row of a data frame `.df`, with optional parallelism and progress bar.
#' Each row is converted to a named list before being passed to `.f`, enabling flexible access to variables by name.
#'
#' @param .df A data frame whose rows will be iterated over.
#' @param .f A function applied to each row, which receives a named list.
#' @param ncores Integer. Number of cores to use for parallel processing. Default is `NULL` (sequential).
#' @param pb Logical. Whether to display a progress bar. Default is `FALSE`.
#' @param ... Additional arguments passed to `.f`.
#'
#' @return A list of results returned by applying `.f` to each row as a list.
#'
#' @examples
#' df <- data.frame(name = c("Mister", "Hipster"), age = c(30, 25))
#'
#' # Create personalized messages
#' fmapr(df, function(row) paste(row$name, "is", row$age, "years old"))
#'
#' # Row-wise model formulas
#' formulas <- data.frame(
#'   response = c("y1", "y2"),
#'   predictor = c("x1", "x2"),
#'   stringsAsFactors = FALSE
#' )
#'
#' fmapr(formulas, function(row) {
#'   reformulate(row$predictor, row$response)
#' })
#'
#' @export

fmapr <- function(.df, .f, ncores = NULL, pb = FALSE, ...) {
  .f <- match.fun(.f)
  rows <- split(.df, seq_len(nrow(.df)))
  fapply(rows, function(row) .f(as.list(row), ...), ncores = ncores, pb = pb)
}

#' Apply a function over a list or vector with optional parallelism and progress
#'
#' A lightweight and fast version of `lapply()` with support for multicore (Unix) and snow-style
#' clusters via `parallel`, with internal progress bar tracking and message suppression.
#' When `pb = TRUE`, progress advances when individual tasks complete, including in the
#' parallel paths, so counts reflect completed jobs rather than internal chunk boundaries.
#' Progress rendering is throttled for large workloads to keep console redraw overhead bounded,
#' while status output reports elapsed time and estimated time remaining.
#'
#' @param .x A list or atomic vector.
#' @param .f Function to apply.
#' @param ncores Number of cores to use (default: 1 = sequential).
#' @param pb Show progress bar? (default: FALSE). When enabled, progress is driven by
#' completed tasks. In parallel mode, updates occur as workers return results.
#' @param cl A cluster object (from parallel::makeCluster), or integer for core count.
#' @param load_balancing Logical. Use `parLapplyLB` if `TRUE` (default: `FALSE`).
#' @param ... Additional arguments passed to `.f`.
#'
#' @return A list of results.
#'
#' @examples
#' # Basic usage (sequential)
#' fapply(1:5, sqrt)
#'
#' # With progress bar (sequential)
#' fapply(1:5, function(x) { Sys.sleep(0.1); x^2 }, pb = TRUE)
#'
#' # Multicore on Unix (if available)
#' \donttest{
#' if (.Platform$OS.type != "windows") {
#'   fapply(1:10, sqrt, ncores = 2)
#' }
#' }
#'
#' # With user-created cluster (portable across platforms)
#' \donttest{
#' cl <- parallel::makeCluster(2)
#' fapply(1:10, sqrt, cl = cl)
#' parallel::stopCluster(cl)
#' }
#'
#' # Heavy computation example with exact completion-driven progress
#' \donttest{
#' heavy_fn <- function(x) { Sys.sleep(0.05); x^2 }
#' fapply(1:20, heavy_fn, ncores = 2, pb = TRUE)
#' }
#'
#' @export

fapply <- function(.x, .f, ncores = 1, pb = FALSE, cl = NULL, load_balancing = TRUE, ...) {
  .f <- match.fun(.f)
  if (!is.vector(.x) || is.object(.x)) .x <- as.list(.x)
  if (!length(.x)) return(list())

  is_windows <- .Platform$OS.type == "windows"
  use_parallel <- isTRUE(ncores > 1L)

  # disable crashpad messages
  Sys.setenv(CHROME_CRASHPAD_PIPE_NAME = "disable")

  # sequential fallback
  if (!use_parallel && is.null(cl)) {
    if (pb) {
      pb_bar <- functionals_progress_bar(min = 0, max = length(.x))
      on.exit(pb_bar$kill(), add = TRUE)
      out <- vector("list", length(.x))
      for (i in seq_along(.x)) {
        captured_messages <- character()
        result <- withCallingHandlers(
          capture.output(
            out[[i]] <- .f(.x[[i]], ...),
            type = "output"
          ),
          message = function(m) {
            captured_messages <<- c(captured_messages, conditionMessage(m))
            invokeRestart("muffleMessage")
          }
        )
        emitted_text <- c(captured_messages, result)
        if (length(emitted_text)) {
          if (i < length(.x)) {
            pb_bar$up(i)
            pb_bar$emit(emitted_text)
          } else {
            pb_bar$emit(emitted_text, redraw = FALSE)
            pb_bar$up(i)
          }
        } else {
          pb_bar$up(i)
        }
      }
      return(out)
    } else {
      return(lapply(.x, .f, ...))
    }
  }

  # if user passed a cluster
  if (inherits(cl, "cluster")) {
    PAR_FUN <- if (load_balancing) parallel::parLapplyLB else parallel::parLapply
    if (pb) {
      pb_bar <- functionals_progress_bar(min = 0, max = length(.x))
      on.exit(pb_bar$kill(), add = TRUE)
      task_fun <- function(item) {
        tryCatch(
          list(ok = TRUE, value = .f(item, ...)),
          error = function(e) list(ok = FALSE, message = conditionMessage(e), call = conditionCall(e))
        )
      }
      return(.functionals_cluster_queue(cl, .x, task_fun, pb_bar = pb_bar))
    } else {
      return(PAR_FUN(cl, .x, .f, ...))
    }
  }

  # if user requested multicore (Unix only)
  if (!is_windows && is.null(cl)) {
    if (pb) {
      pb_bar <- functionals_progress_bar(min = 0, max = length(.x))
      on.exit(pb_bar$kill(), add = TRUE)
      task_fun <- function(item) {
        tryCatch(
          list(
            ok = TRUE,
            value = suppressWarnings(suppressMessages(.f(item, ...)))
          ),
          error = function(e) list(ok = FALSE, message = conditionMessage(e), call = conditionCall(e))
        )
      }
      return(.functionals_multicore_queue(.x, task_fun, ncores = ncores, pb_bar = pb_bar))
    } else {
      return(
        suppressWarnings(suppressMessages(
          parallel::mclapply(.x, .f, ..., mc.cores = ncores, mc.silent = TRUE)
        ))
      )
    }
  }

  # Windows or fallback with no cluster -> create PSOCK cluster
  cl <- parallel::makeCluster(ncores)
  on.exit(parallel::stopCluster(cl), add = TRUE)

  PAR_FUN <- if (load_balancing) parallel::parLapplyLB else parallel::parLapply
  if (pb) {
    pb_bar <- functionals_progress_bar(min = 0, max = length(.x))
    on.exit(pb_bar$kill(), add = TRUE)
    task_fun <- function(item) {
      tryCatch(
        list(ok = TRUE, value = .f(item, ...)),
        error = function(e) list(ok = FALSE, message = conditionMessage(e), call = conditionCall(e))
      )
    }
    return(.functionals_cluster_queue(cl, .x, task_fun, pb_bar = pb_bar))
  } else {
    return(PAR_FUN(cl, .x, .f, ...))
  }
}

library(functionals)

capture_progress_output <- function(expr) {
  paste(capture.output(force(expr), type = "output"), collapse = "\n")
}

progress_lines <- function(output) {
  lines <- strsplit(gsub("\r", "\n", output, fixed = TRUE), "\n", fixed = TRUE)[[1]]
  lines[nzchar(trimws(lines))]
}

test_that("fapply stays silent when progress is disabled", {
  output <- capture_progress_output({
    out <- fapply(1:3, identity, pb = FALSE)
  })

  expect_identical(output, "")
  expect_equal(unname(out), as.list(1:3))
})

test_that("sequential progress uses task totals and reports ETA from completed work", {
  skip_on_cran()

  output <- capture_progress_output({
    out <- fapply(1:4, function(x) {
      Sys.sleep(1.05)
      x
    }, pb = TRUE)
  })

  lines <- progress_lines(output)
  counts <- c("0/4", "1/4", "2/4", "3/4", "4/4")

  expect_true(all(vapply(counts, function(count) any(grepl(count, lines, fixed = TRUE)), logical(1))))
  expect_true(any(grepl("elapsed", lines, fixed = TRUE)))
  expect_true(any(grepl("eta", lines[grepl("1/4", lines, fixed = TRUE)], fixed = TRUE)))
  expect_true(any(grepl("eta", lines, fixed = TRUE)))
})

test_that("cluster-backed progress counts completed tasks rather than chunks", {
  skip_on_cran()

  cl <- parallel::makeCluster(2)
  on.exit(parallel::stopCluster(cl), add = TRUE)

  output <- capture_progress_output({
    out <- fapply(as.list(1:4), function(x) {
      Sys.sleep(c(0.12, 0.24, 0.36, 0.48)[[x]])
      x
    }, cl = cl, pb = TRUE)
  })

  lines <- progress_lines(output)
  counts <- c("1/4", "2/4", "3/4", "4/4")

  expect_true(all(vapply(counts, function(count) any(grepl(count, lines, fixed = TRUE)), logical(1))))
  expect_false(any(grepl("2/2", lines, fixed = TRUE)))
  expect_equal(unname(out), as.list(1:4))
})

test_that("multicore progress counts completed tasks rather than chunks", {
  skip_on_cran()
  skip_on_os("windows")

  output <- capture_progress_output({
    out <- fapply(as.list(1:4), function(x) {
      Sys.sleep(c(0.12, 0.24, 0.36, 0.48)[[x]])
      x
    }, ncores = 2, pb = TRUE)
  })

  lines <- progress_lines(output)
  counts <- c("1/4", "2/4", "3/4", "4/4")

  expect_true(all(vapply(counts, function(count) any(grepl(count, lines, fixed = TRUE)), logical(1))))
  expect_false(any(grepl("2/2", lines, fixed = TRUE)))
  expect_equal(unname(out), as.list(1:4))
})

test_that("messages do not produce duplicate completed progress lines", {
  skip_on_cran()

  output <- capture_progress_output({
    out <- fapply(1:2, function(x) {
      message(sprintf("msg-%s", x))
      x
    }, pb = TRUE)
  })

  lines <- progress_lines(output)
  completed_lines <- lines[grepl("100% 2/2", lines, fixed = TRUE)]

  expect_lte(length(completed_lines), 1)
  expect_equal(unname(out), as.list(1:2))
})

test_that("cluster-backed progress propagates task errors", {
  skip_on_cran()

  cl <- parallel::makeCluster(2)
  on.exit(parallel::stopCluster(cl), add = TRUE)

  expect_error(
    fapply(as.list(1:4), function(x) {
      if (x == 3) stop("cluster-boom")
      x
    }, cl = cl, pb = TRUE),
    "cluster-boom"
  )
})

test_that("cluster-backed progress refreshes elapsed time while waiting", {
  skip_on_cran()

  cl <- parallel::makeCluster(2)
  on.exit(parallel::stopCluster(cl), add = TRUE)

  output <- capture_progress_output({
    out <- fapply(as.list(1:2), function(x) {
      Sys.sleep(c(1.2, 1.6)[[x]])
      x
    }, cl = cl, pb = TRUE)
  })

  lines <- progress_lines(output)

  expect_true(any(grepl("0/2 elapsed 00:01", lines, fixed = TRUE)))
  expect_equal(unname(out), as.list(1:2))
})

test_that("multicore progress propagates task errors", {
  skip_on_cran()
  skip_on_os("windows")

  expect_error(
    fapply(as.list(1:4), function(x) {
      if (x == 3) stop("multicore-boom")
      x
    }, ncores = 2, pb = TRUE),
    "multicore-boom"
  )
})

test_that("multicore progress refreshes elapsed time while waiting", {
  skip_on_cran()
  skip_on_os("windows")

  output <- capture_progress_output({
    out <- fapply(as.list(1:2), function(x) {
      Sys.sleep(c(1.2, 1.6)[[x]])
      x
    }, ncores = 2, pb = TRUE)
  })

  lines <- progress_lines(output)

  expect_true(any(grepl("0/2 elapsed 00:01", lines, fixed = TRUE)))
  expect_equal(unname(out), as.list(1:2))
})

test_that("fapply works sequentially", {
  x <- 1:5
  expect_equal(unname(fapply(x, function(x) x^2)), as.list(x^2))
})

test_that("fapply works with progress bar and parallel", {
  skip_on_cran()
  x <- 1:10
  res <- fapply(x, function(x) x + 1, ncores = 2, pb = TRUE)
  expect_equal(unname(res), as.list(x + 1))
})

test_that("fapply handles empty input", {
  expect_equal(fapply(list(), identity), list())
})

#' Walk over a vector or list with side effects
#'
#' Applies a function `.f` to each element of `.x`, typically for its side effects (e.g., printing, writing files).
#' This function is the side-effect-friendly equivalent of `fmap()`. Supports parallel execution and progress bar display.
#'
#' @param .x A list or atomic vector of elements to iterate over.
#' @param .f A function to apply to each element of `.x`. Should be called primarily for side effects.
#' @param ncores Integer. Number of cores to use for parallel processing. Default is `NULL` (sequential).
#' @param pb Logical. Whether to show a progress bar. Default is `FALSE`.
#' @param ... Additional arguments passed to `.f`.
#'
#' @return Invisibly returns `.x`, like `purrr::walk()`.
#'
#' @examples
#' # Print each element
#' fwalk(1:3, print)
#'
#' # Simulate writing files in parallel
#' \donttest{
#' fwalk(1:3, function(i) {
#'   cat(paste("Processing item", i, "\n"))
#'   Sys.sleep(0.5)
#' }, ncores = 2, pb = TRUE)
#' }
#'
#' @export

fwalk <- function(.x, .f, ncores = NULL, pb = FALSE, ...) {
  .f <- match.fun(.f)
  tmp <- fapply(.x, function(x) {.f(x, ...); NULL}, ncores = ncores, pb = pb)
  invisible(.x)
}

test_that("fwalk executes function for side effects", {
  calls <- character()
  fwalk(1:3, function(x) calls <<- c(calls, paste0("call", x)))
  expect_equal(calls, c("call1", "call2", "call3"))
})

test_that("fwalk returns input invisibly", {
  expect_invisible(fwalk(1:3, function(x) NULL))
})


test_that("fwalk returns input and does so invisibly", {
  out <- fwalk(1:3, function(x) NULL)
  expect_equal(out, 1:3)
  expect_invisible(fwalk(1:3, function(x) NULL))
})


test_that("fwalk supports parallel execution with progress", {
  skip_on_cran()
  log <- tempfile()
  fwalk(1:4, function(i) {
    cat(paste("Task", i), file = log, append = TRUE, sep = "\n")
  }, ncores = 2, pb = TRUE)

  lines <- readLines(log)
  expect_length(lines, 4)
  expect_true(all(grepl("Task", lines)))
})

#' Compose multiple functions
#'
#' Create a new function by composing several functions, applied from right to left.
#' Equivalent to `f1(f2(f3(...)))`.
#'
#' @param ... Functions to compose. Each must take a single argument and return an object
#' compatible with the next function in the chain.
#'
#' @return A new function equivalent to nested application of the input functions.
#' @examples
#' square <- function(x) x^2
#' add1 <- function(x) x + 1
#'
#' f <- fcompose(sqrt, square, add1)  # => sqrt(square(x + 1))
#' f(4)  # => sqrt((4 + 1)^2) = sqrt(25) = 5
#'
#' # More compact
#' fcompose(log, exp)(2)  # log(exp(2)) = 2
#'
#' @export
fcompose <- function(...) {
  funcs <- list(...)
  if (length(funcs) == 0) stop("At least one function must be provided.")
  function(x) {
    for (f in rev(funcs)) {
      f <- match.fun(f)
      x <- f(x)
    }
    x
  }
}

test_that("fcompose works with numeric functions", {
  f <- fcompose(sqrt, function(x) x^2, function(x) x + 1)
  expect_equal(f(4), 5)
})

test_that("fcompose fails with no input functions", {
  expect_error(fcompose(), "At least one function must be provided.")
})


test_that("fcompose returns original function if one is passed", {
  f <- fcompose(function(x) x + 1)
  expect_equal(f(3), 4)
})

#' Functional mapping with optional parallelism and progress bars
#'
#' Applies a function `.f` to each element of `.x`, with optional parallel processing and progress bar support.
#' Progress behavior is inherited from `fapply()`, including completion-driven updates
#' in the parallel execution paths, bounded redraw frequency for large workloads,
#' and elapsed/ETA status reporting.
#'
#' @param .x A list or atomic vector of elements to iterate over.
#' @param .f A function to apply to each element of `.x`. Can be a function or a string naming a function.
#' @param ncores Integer. Number of CPU cores to use for parallel processing. Default is `NULL` (sequential).
#' @param pb Logical. Whether to show a progress bar. Default is `FALSE`. In parallel
#' mode, progress updates as each completed task returns.
#' @param ... Additional arguments passed to `.f`.
#'
#' @return A list of results, one for each element of `.x`.
#'
#' @examples
#' slow_fn <- function(x) { Sys.sleep(0.01); x^2 }
#' x <- 1:100
#'
#' # Basic usage
#' fmap(x, slow_fn)
#'
#' # With progress bar
#' fmap(x, slow_fn, pb = TRUE)
#'
#' # With parallel execution (non-Windows)
#' \donttest{
#' if (.Platform$OS.type != "windows") {
#'   fmap(x, slow_fn, ncores = 2, pb = TRUE)
#' }
#' }
#'
#' @export
fmap <- function(.x, .f, ncores = NULL, pb = FALSE, ...) {
  .f <- match.fun(.f)
  fapply(.x, .f, ncores = ncores %||% 1, pb = pb, ...)
}

test_that("fmap works like fapply", {
  x <- 1:5
  expect_equal(unname(fmap(x, function(x) x + 1)), as.list(2:6))
})

test_that("fmap works with progress and parallel", {
  skip_on_cran()
  x <- 1:5
  res <- fmap(x, function(x) x * 10, ncores = 2, pb = TRUE)
  expect_equal(unname(res), as.list(x * 10))
})

test_that("fmap returns empty list on empty input", {
  expect_equal(fmap(list(), identity), list())
})

#' Functional reduce
#'
#' Apply a binary function iteratively over a list or vector, reducing it to a single value or a sequence of intermediate results.
#' This is a wrapper around [Reduce()] that supports optional initial values, right-to-left evaluation,
#' accumulation of intermediate steps, and output simplification.
#'
#' @param .x A vector or list to reduce.
#' @param .f A binary function to apply. Can be given as a function or quoted (e.g., `\`+\``).
#' @param .init Optional initial value passed to [Reduce()]. If `NULL`, reduction starts from the first two elements.
#' @param .right Logical. If `TRUE`, reduction is performed from right to left.
#' @param .accumulate Logical. If `TRUE`, returns a list of intermediate results (like a scan).
#' @param .simplify Logical. If `TRUE` and all intermediate results are length 1, the output is simplified to a vector.
#'
#' @return A single value (default) or a list/vector of intermediate results if `.accumulate = TRUE`.
#'
#' @examples
#' freduce(1:5, `+`)                          # => 15
#' freduce(letters[1:4], paste0)             # => "abcd"
#' freduce(list(1, 2, 3), `*`)               # => 6
#' freduce(1:3, `+`, .init = 10)             # => 16
#' freduce(1:3, paste0, .right = TRUE)       # => "321"
#' freduce(1:4, `+`, .accumulate = TRUE)     # => c(1, 3, 6, 10)
#'
#' @export
freduce <- function(.x, .f, .init = NULL, .right = FALSE, .accumulate = FALSE, .simplify = TRUE) {
  .f <- match.fun(.f)
  if (!is.list(.x)) .x <- as.list(.x)
  if (is.null(.init)) {
    Reduce(.f, .x, right = .right, accumulate = .accumulate, simplify = .simplify)
  } else {
    Reduce(.f, .x, init = .init, right = .right, accumulate = .accumulate, simplify = .simplify)
  }
}

test_that("freduce sums numbers correctly", {
  expect_equal(freduce(1:5, `+`), 15)
})

test_that("freduce concatenates strings", {
  expect_equal(freduce(letters[1:4], paste0), "abcd")
})

test_that("freduce works with a custom binary function", {
  out <- freduce(list(1, 2, 3), function(x, y) x * y)
  expect_equal(out, 6)
})

test_that("freduce handles init argument", {
  expect_equal(freduce(1:3, `+`, .init = 10), 16)
})


test_that("freduce supports right-to-left reduction", {
  expect_equal(freduce(1:3, function(x, y) paste0(x, y), .right = TRUE), "123")
})

test_that("freduce supports accumulation and simplification", {
  expect_equal(freduce(1:4, `+`, .accumulate = TRUE), c(1, 3, 6, 10))
  expect_equal(freduce(1:4, `+`, .accumulate = TRUE, .simplify = FALSE), as.list(c(1, 3, 6, 10)))
})

#' Apply a function over multiple argument lists in parallel
#'
#' Applies a function `.f` over multiple aligned lists in `.l`.
#' Each element of `.l` should be a list or vector of the same length. Each call to `.f` receives one element from each list.
#' Supports parallel execution and progress display.
#'
#' @param .l A list of vectors or lists. All elements must be of equal length.
#' @param .f A function to apply. It must accept as many arguments as there are elements in `.l`.
#' @param ncores Integer. Number of cores to use for parallel processing. Default is `NULL` (sequential).
#' @param pb Logical. Whether to display a progress bar. Default is `FALSE`.
#' @param ... Additional arguments passed to `.f`.
#'
#' @return A list of results obtained by applying `.f` to each tuple from `.l`.
#'
#' @examples
#' # Fit a linear model for each response variable using the same predictor
#' df <- data.frame(
#'   y1 = rnorm(100),
#'   y2 = rnorm(100),
#'   x = rnorm(100)
#' )
#'
#' # List of formulas and data
#' formulas <- list(y1 ~ x, y2 ~ x)
#' data_list <- list(df, df)
#'
#' fmapn(list(formula = formulas, data = data_list),
#'       function(formula, data) lm(formula, data = data))
#'
#' # Extract model summaries in parallel
#' models <- fmapn(
#'   list(formula = formulas, data = data_list),
#'   function(formula, data) summary(lm(formula, data = data))$r.squared
#' )
#'
#' @export

fmapn <- function(.l, .f, ncores = NULL, pb = FALSE, ...) {
  .f <- match.fun(.f)
  args <- do.call(Map, c(f = list(function(...) list(...)), .l))
  fapply(args, function(x) do.call(.f, x), ncores = ncores, pb = pb, ...)
}

test_that("fmapn applies function over multiple lists", {
  x <- list(1:3, 4:6)
  out <- fmapn(x, function(a, b) a + b)

  expect_equal(unname(out), as.list(c(5, 7, 9)))
})

test_that("fmapn handles character and numeric lists", {
  first <- letters[1:3]
  second <- 1:3
  out <- fmapn(list(first, second), function(a, b) paste0(a, b))

  expect_equal(unname(out), as.list(c("a1", "b2", "c3")))
})

test_that("fmapn supports parallel execution with progress", {
  skip_on_cran()
  a <- 1:10
  b <- 11:20

  out <- fmapn(list(a, b), function(x, y) x * y, ncores = 2, pb = TRUE)

  expect_equal(unname(out), as.list(a * b))
})

#' Functional loop with optional parallelism and progress bar
#'
#' `floop()` applies a function `.f` to each element of `.x`, optionally in parallel, and with an optional progress bar.
#' Unlike `fwalk()`, it can return results or be used purely for side effects (like a for-loop).
#'
#' @param .x A vector or list of elements to iterate over.
#' @param .f A function to apply to each element of `.x`.
#' @param ncores Integer. Number of cores to use. Default is 1 (sequential).
#' @param pb Logical. Show a progress bar? Default is `FALSE`.
#' @param .capture Logical. Should results of `.f` be captured and returned? If `FALSE`, acts like a side-effect loop.
#' @param ... Additional arguments passed to `.f`.
#'
#' @return A list of results if `.capture = TRUE`, otherwise returns `.x` invisibly.
#'
#' @examples
#' # Functional loop that collects output
#' floop(1:3, function(i) i^2)
#'
#' # Side-effect only loop (like for-loop with cat)
#' \donttest{
#' floop(1:5, function(i) cat(" Processing", i, "\n"),
#'       pb = TRUE, .capture = FALSE)
#' }
#'
#' @export

floop <- function(.x, .f, ncores = 1, pb = FALSE, .capture = TRUE, ...) {
  .f <- match.fun(.f)
  if (!is.vector(.x) || is.object(.x)) .x <- as.list(.x)
  fwrap <- function(x) {.f(x, ...)}
  result <- fapply(.x, fwrap, ncores = ncores, pb = pb)
  if (.capture) result else invisible(.x)
}

test_that("floop returns expected results", {
  out <- floop(1:3, function(x) x + 1)
  expect_equal(unname(out), list(2, 3, 4))
})

test_that("floop respects .capture = FALSE", {
  calls <- character()
  invisible(floop(1:3, function(i) calls <<- c(calls, paste0("log", i)), .capture = FALSE))
  expect_equal(calls, c("log1", "log2", "log3"))
})

test_that("floop works with ncores > 1", {
  out <- floop(1:4, function(x) x^2, ncores = 2)
  expect_equal(unname(out), list(1, 4, 9, 16))
})

#' @importFrom utils flush.console capture.output
NULL


#' Validate and normalize fapply arguments
#'
#' Internal helper to check and coerce inputs for `fapply()`.
#'
#' @keywords internal
#' @noRd
.check_fapply_args <- function(.x, .f, ncores, pb) {
  # check .x is atomic or list-like
  if (!is.vector(.x) || is.object(.x)) .x <- as.list(.x)

  # empty input warning
  if (length(.x) == 0L) {
    warning("Input `.x` is empty. Returning empty list.")
    return(list(result = list(), .x = .x))
  }

  # check .f
  .f <- match.fun(.f)

  # check ncores
  if (!is.null(ncores)) {
    if (!is.numeric(ncores) || ncores < 1 || is.na(ncores)) {
      warning("`ncores` must be a positive integer. Defaulting to sequential.")
      ncores <- 1
    }
  } else {
    ncores <- 1
  }

  # windows fallback
  if (.Platform$OS.type == "windows" && ncores > 1) {
    warning("Parallel execution on Windows uses clusters. Falling back to sequential for consistency.")
    ncores <- 1
  }

  # check pb
  if (!is.logical(pb) || length(pb) != 1) {
    warning("`pb` must be TRUE or FALSE. Disabling progress bar.")
    pb <- FALSE
  }

  list(result = NULL, .x = .x, .f = .f, ncores = ncores, pb = pb)
}


#' Create a custom console progress bar
#'
#' Internal utility to build a simple console-based progress bar with estimated time remaining.
#' Used inside `fapply()` when `pb = TRUE`.
#'
#' @keywords internal
#' @noRd
functionals_progress_bar <- function(min = 0, max = 1, style = 1, width = NA, char = "=") {
  start_time <- proc.time()[["elapsed"]]
  last_draw <- -Inf
  last_value <- min
  i <- min
  if (is.na(width)) width <- getOption("width")
  total <- max - min
  redraw_step <- max(1L, ceiling(max(total, 1) / 100))
  tick_interval <- 0.25

  get_time_str <- function(seconds) {
    seconds <- max(0, round(seconds))
    minutes <- seconds %/% 60
    sec <- seconds %% 60
    sprintf("%02d:%02d", minutes, sec)
  }

  draw <- function(value, now) {
    percent <- if (total > 0) (value - min) / total else 1
    percent <- max(0, min(1, percent))
    completed <- value - min
    total_steps <- max(total, 0)
    elapsed <- now - start_time
    eta <- if (completed > 0 && percent > 0) elapsed * (1 - percent) / percent else NA
    status <- sprintf(
      "%3d%% %d/%d elapsed %s",
      round(percent * 100),
      completed,
      total_steps,
      get_time_str(elapsed)
    )
    if (!is.na(eta)) status <- sprintf("%s eta %s", status, get_time_str(eta))

    bar_width <- max(10, min(30, width - nchar(status) - 6))
    done <- floor(bar_width * percent)
    left <- bar_width - done
    clear()
    cat(sprintf("\r[%s%s] %s", strrep(char, done), strrep(" ", left), status))
    flush.console()
    last_draw <<- now
    last_value <<- value
  }

  update <- function(value) {
    if (!is.finite(value) || value < min || value > max) return()
    i <<- value
    now <- proc.time()[["elapsed"]]
    is_final <- isTRUE(value >= max)
    advanced_enough <- (value - last_value) >= redraw_step
    is_initial <- !is.finite(last_draw)
    if (is_initial || ((now - last_draw) >= 0.1 && advanced_enough) || is_final) {
      draw(value, now)
    }
  }

  tick <- function() {
    if (i >= max) return()
    now <- proc.time()[["elapsed"]]
    if ((now - last_draw) >= tick_interval) {
      draw(i, now)
    }
  }

  clear <- function() {
    cat("\r", strrep(" ", width), "\r", sep = "")
  }

  emit <- function(text, redraw = TRUE) {
    if (!length(text)) return()
    clear()
    cat(paste(text, collapse = "\n"))
    if (!grepl("\n$", text[[length(text)]])) cat("\n")
    if (redraw) draw(i, proc.time()[["elapsed"]])
  }

  kill <- function() cat("\n")
  update(i)
  list(up = update, tick = tick, kill = kill, emit = emit, clear = clear)
}


#' Split indices for chunked parallel processing
#'
#' Helper function to divide a sequence of indices into balanced chunks for parallel processing.
#' Used internally by `fapply()` to manage load balancing.
#'
#' @param nx Integer. Total number of elements to split.
#' @param ncl Integer. Number of cores (chunks).
#' @param nout Approximate number of output groups for progress bar feedback.
#'
#' @return A list of integer vectors, each representing a chunk of indices.
#'
#' @keywords internal
#' @noRd
splitpb <- function(nx, ncl, nout = 100) {
  i <- seq_len(nx)
  if (ncl == 0L) return(list())
  k <- max(1L, ceiling(ceiling(nx / ncl) / nout))
  g <- 1L + (i - 1L) %/% as.integer(ncl * k)
  structure(split(i, g), names = NULL)
}


#' Unpack a task result emitted by a worker wrapper
#'
#' @keywords internal
#' @noRd
.functionals_unwrap_task_result <- function(result) {
  if (is.list(result) && identical(result$ok, TRUE)) {
    return(result$value)
  }

  if (is.list(result) && identical(result$ok, FALSE)) {
    stop(result$message, call. = FALSE)
  }

  if (inherits(result, "try-error")) {
    stop(as.character(result), call. = FALSE)
  }

  result
}


#' Execute tasks on a cluster with exact completion-driven progress
#'
#' @keywords internal
#' @noRd
.functionals_cluster_queue <- function(cl, .x, task_fun, pb_bar = NULL) {
  if (!length(.x)) return(list())

  parallel_ns <- asNamespace("parallel")
  send_call <- get("sendCall", envir = parallel_ns)
  recv_one_result <- get("recvOneResult", envir = parallel_ns)
  is_sock_cluster <- inherits(cl, "SOCKcluster") || all(vapply(cl, inherits, logical(1), what = "SOCKnode"))

  out <- vector("list", length(.x))
  next_idx <- 1L
  n_workers <- min(length(cl), length(.x))
  socklist <- if (is_sock_cluster) lapply(cl, function(x) x$con) else NULL

  for (worker_idx in seq_len(n_workers)) {
    send_call(cl[[worker_idx]], task_fun, list(.x[[next_idx]]), tag = next_idx)
    next_idx <- next_idx + 1L
  }

  completed <- 0L
  while (completed < length(.x)) {
    res <- if (is_sock_cluster) {
      ready <- socketSelect(socklist, timeout = 0.25)
      if (!any(ready)) {
        if (!is.null(pb_bar)) pb_bar$tick()
        next
      }
      node <- which.max(ready)
      value <- unserialize(socklist[[node]])
      list(value = value$value, node = node, tag = value$tag)
    } else {
      recv_one_result(cl)
    }

    out[[res$tag]] <- .functionals_unwrap_task_result(res$value)
    completed <- completed + 1L
    if (!is.null(pb_bar)) pb_bar$up(completed)

    if (next_idx <= length(.x)) {
      send_call(cl[[res$node]], task_fun, list(.x[[next_idx]]), tag = next_idx)
      next_idx <- next_idx + 1L
    }
  }

  out
}


#' Execute tasks with mcparallel and exact completion-driven progress
#'
#' @keywords internal
#' @noRd
.functionals_multicore_queue <- function(.x, task_fun, ncores, pb_bar = NULL) {
  if (!length(.x)) return(list())

  parallel_ns <- asNamespace("parallel")
  cleanup <- get("cleanup", envir = parallel_ns)
  on.exit(cleanup(kill = TRUE, detach = FALSE), add = TRUE)

  out <- vector("list", length(.x))
  jobs <- list()
  next_idx <- 1L
  max_jobs <- min(ncores, length(.x))

  launch_job <- function(idx) {
    parallel::mcparallel(task_fun(.x[[idx]]), name = as.character(idx), silent = TRUE)
  }

  while (length(jobs) < max_jobs && next_idx <= length(.x)) {
    jobs[[as.character(next_idx)]] <- launch_job(next_idx)
    next_idx <- next_idx + 1L
  }

  completed <- 0L
  while (completed < length(.x)) {
    res <- parallel::mccollect(jobs, wait = FALSE, timeout = 0.1)
    if (is.null(res)) {
      if (!is.null(pb_bar)) pb_bar$tick()
      next
    }

    for (name in names(res)) {
      out[[as.integer(name)]] <- .functionals_unwrap_task_result(res[[name]])
      jobs[[name]] <- NULL
      completed <- completed + 1L
      if (!is.null(pb_bar)) pb_bar$up(completed)

      if (next_idx <= length(.x)) {
        jobs[[as.character(next_idx)]] <- launch_job(next_idx)
        next_idx <- next_idx + 1L
      }
    }
  }

  out
}

test_that("check_fapply_args handles invalid ncores and pb", {
  expect_warning(res <- functionals:::.check_fapply_args(1:3, sum, ncores = -1, pb = "maybe"))
  expect_equal(res$ncores, 1)
  expect_false(res$pb)
})

#' functionals: Functional programming with parallelism and progress track in R
#'
#' This package provides a consistent and intuitive API for applying functions
#' over lists, vectors, and data frames, with built-in support for parallelism
#' and progress bars.
#'
#' @docType package
#' @keywords internal
"_PACKAGE"

#' Apply a function column-wise with name access and parallelism
#'
#' Applies a function `.f` to each column of a data frame `.df`. Each call receives both
#' the column vector and its name, enabling name-aware column processing. Supports parallel execution and progress display.
#'
#' @param .df A data frame whose columns will be iterated over.
#' @param .f A function that takes two arguments: the column vector and its name.
#' @param ncores Integer. Number of cores to use for parallel processing. Default is `NULL` (sequential).
#' @param pb Logical. Whether to display a progress bar. Default is `FALSE`.
#' @param ... Additional arguments passed to `.f`.
#'
#' @return A list of results obtained by applying `.f` to each column of `.df`.
#'
#' @examples
#' df <- data.frame(a = 1:3, b = 4:6)
#'
#' # Apply a function that returns column mean and name
#' fmapc(df, function(x, name) {
#'   list(mean = mean(x), var = var(x), name = name)
#' })
#'
#' # With progress and parallel execution
#' \donttest{
#' fmapc(df, function(x, name) mean(x), ncores = 2, pb = TRUE)
#' }
#'
#' @export

fmapc <- function(.df, .f, ncores = NULL, pb = FALSE, ...) {
  .f <- match.fun(.f)
  cols <- as.list(.df)
  names(cols) <- names(.df)
  col_pairs <- mapply(function(x, n) list(x, n), cols, names(cols), SIMPLIFY = FALSE)
  fapply(col_pairs, function(pair) .f(pair[[1]], pair[[2]], ...), ncores = ncores, pb = pb)
}

test_that("fmapc applies function to each column with column names", {
  df <- data.frame(a = 1:3, b = 4:6)
  out <- fmapc(df, function(x, name) paste0(name, ":", sum(x)))

  expect_length(out, ncol(df))
  expect_true(all(sapply(out, is.character)))
})

test_that("fmapc supports numeric summaries with different output", {
  df <- data.frame(a = 1:10, b = rnorm(10))
  out <- fmapc(df, function(x, name) mean(x))

  expect_length(out, 2)
  expect_true(all(sapply(out, is.numeric)))
})

test_that("fmapc supports parallel execution", {
  skip_on_cran()
  df <- data.frame(a = 1:100, b = 101:200)
  out <- fmapc(df, function(x, name) mean(x), ncores = 2, pb = TRUE)

  expect_equal(length(out), 2)
  expect_true(all(sapply(out, is.numeric)))
})

#' Repeat an expression or function call multiple times
#'
#' Repeats an expression or function evaluation `times` times. If `expr` is a function, it is invoked
#' with optional input `.x` and additional arguments. If `expr` is a quoted expression, it is evaluated
#' in the parent environment. Supports parallel processing and optional simplification of results.
#'
#' @param .x Optional input passed to `expr` if `expr` is a function. Default is `NULL`.
#' @param times Integer. Number of repetitions.
#' @param expr A function or an unevaluated expression to repeat. If a function, it will be called `times` times.
#' @param simplify Logical. If `TRUE`, attempts to simplify the result using `simplify2array()`. Default is `FALSE`.
#' @param ncores Integer. Number of cores to use for parallel execution. Default is `NULL` (sequential).
#' @param pb Logical. Whether to display a progress bar. Default is `FALSE`.
#' @param ... Additional arguments passed to the function `expr` if it is callable.
#'
#' @return A list of outputs (or a simplified array if `simplify = TRUE`) from evaluating `expr` multiple times.
#'
#' @examples
#' # Repeat a pure function call
#' frepeat(times = 3, expr = function() rnorm(1))
#'
#' # Repeat a function with input `.x`
#' frepeat(.x = 10, times = 3, expr = function(x) rnorm(1, mean = x))
#'
#' # Repeat an unevaluated expression (evaluated with `eval()`)
#' frepeat(times = 2, expr = quote(rnorm(1)))
#'
#' # Simplify the output to an array
#' frepeat(times = 3, expr = function() rnorm(1), simplify = TRUE)
#'
#' # Monte Carlo simulation: estimate coverage of a 95% CI for sample mean
#' mc_result <- frepeat(
#'   times = 1000, simplify = TRUE, pb = TRUE, ncores = 1,
#'   expr = function() {
#'     sample <- rnorm(30, mean = 0, sd = 1)
#'     ci <- t.test(sample)$conf.int
#'     mean(ci[1] <= 0 & 0 <= ci[2])  # true mean inside the interval?
#'   }
#' )
#' mean(mc_result)  # estimated coverage
#'
#' @note If `expr` is passed as a function call (not a function or quoted expression),
#' it will be evaluated immediately, not repeated. Use \code{function(...) \\{ ... \\}} or \code{quote(...)} instead.
#'
#' @export

frepeat <- function(.x = NULL, times, expr, simplify = FALSE, ncores = NULL, pb = FALSE, ...) {
  out <- if (is.function(expr)) {
    fapply(
      seq_len(times),
      function(i) {
        if (is.null(.x)) expr(...) else expr(.x, ...)
      },
      ncores = ncores,
      pb = pb
    )
  } else {
    fapply(seq_len(times), function(i) eval(expr), ncores = ncores, pb = pb)
  }
  if (simplify) simplify2array(out) else out
}

test_that("frepeat evaluates an expression repeatedly", {
  out <- frepeat(times = 3, expr = quote(rnorm(1)))
  expect_length(out, 3)
  expect_true(all(sapply(out, is.numeric)))
})

test_that("frepeat works with a function and optional .x", {
  out1 <- frepeat(times = 3, expr = function() 1)
  expect_equal(unname(out1), list(1, 1, 1))

  out2 <- frepeat(.x = 10, times = 3, expr = function(x) x + 1)
  expect_equal(unname(out2), list(11, 11, 11))
})

test_that("frepeat can simplify the result", {
  out <- frepeat(times = 3, expr = function() 1, simplify = TRUE)
  expect_equal(out, rep(1, 3))
})

test_that("frepeat works in parallel with progress bar", {
  skip_on_cran()
  out <- frepeat(times = 5, expr = function() 1, ncores = 2, pb = TRUE)
  expect_equal(unname(out), rep(list(1), 5))
})

#' Functional Cross-Validation mapping
#'
#' Applies a user-defined function `.f` to each element of `.splits`, typically from
#' cross-validation objects such as `rsample::vfold_cv()`.
#'
#' @param .splits A list of resample splits (e.g., from `rsample::vfold_cv()`).
#' @param .f A function to apply to each split. Typically expects a single `split` object.
#' @param ncores Integer. Number of cores to use for parallel processing. Default is `NULL` (sequential).
#' @param pb Logical. Whether to display a progress bar. Default is `FALSE`.
#' @param ... Additional arguments passed to `.f`.
#'
#' @return A list of results returned by applying `.f` to each element of `.splits`.
#'
#' @examples
#' if (requireNamespace("rsample", quietly = TRUE)) {
#'   set.seed(123)
#'   cv_splits <- rsample::vfold_cv(mtcars, v = 5)
#'
#'   # Apply summary over training sets
#'   fcv(cv_splits$splits, function(split) {
#'     summary(rsample::analysis(split))
#'   })
#'
#'   # With progress and parallel execution
#'   \donttest{
#'     fcv(cv_splits$splits, function(split) {
#'       summary(rsample::analysis(split))
#'     }, ncores = 2, pb = TRUE)
#'   }
#' }
#'
#' @export

fcv <- function(.splits, .f, ncores = NULL, pb = FALSE, ...) {
  .f <- match.fun(.f)
  fapply(.splits, .f, ncores = ncores, pb = pb, ...)
}

test_that("fcv works with rsample::vfold_cv splits", {
  library(rsample)
  set.seed(123)
  cv <- vfold_cv(mtcars, v = 5)

  results <- fcv(cv$splits, function(split) {
    dat <- analysis(split)
    mean(dat$mpg)
  })

  expect_length(results, 5)
  expect_true(all(sapply(results, is.numeric)))
})

test_that("fcv supports parallel execution with progress", {
  skip_on_cran()
  library(rsample)
  cv <- vfold_cv(mtcars, v = 4)

  results <- fcv(cv$splits, function(split) {
    nrow(analysis(split))
  }, ncores = 2, pb = TRUE)

  expect_length(results, 4)
  expect_true(all(sapply(results, is.numeric)))
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fmapg.R
\name{fmapg}
\alias{fmapg}
\title{Apply a function to groups of a data frame in parallel}
\usage{
fmapg(.df, .f, by, ncores = NULL, pb = FALSE, ...)
}
\arguments{
\item{.df}{A data frame to group and apply the function over.}

\item{.f}{A function to apply to each group. The function should accept a data frame (a group).}

\item{by}{A character vector of column names in `.df` used for grouping.}

\item{ncores}{Integer. Number of cores to use for parallel processing. Default is `NULL` (sequential).}

\item{pb}{Logical. Whether to show a progress bar. Default is `FALSE`.}

\item{...}{Additional arguments passed to `.f`.}
}
\value{
A list of results, one for each group defined by `by`.
}
\description{
Applies a function `.f` to each group of rows in a data frame `.df`, where grouping is defined by one or more variables in `by`.
Each group is passed as a data frame to `.f`. Supports parallelism and optional progress display.
}
\examples{
# Group-wise mean of Sepal.Length in iris dataset
fmapg(iris, function(df) mean(df$Sepal.Length), by = "Species")

# Group-wise model fitting with progress and parallelism
\donttest{
fmapg(mtcars, function(df) lm(mpg ~ wt, data = df),
      by = "cyl", ncores = 2, pb = TRUE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fmapn.R
\name{fmapn}
\alias{fmapn}
\title{Apply a function over multiple argument lists in parallel}
\usage{
fmapn(.l, .f, ncores = NULL, pb = FALSE, ...)
}
\arguments{
\item{.l}{A list of vectors or lists. All elements must be of equal length.}

\item{.f}{A function to apply. It must accept as many arguments as there are elements in `.l`.}

\item{ncores}{Integer. Number of cores to use for parallel processing. Default is `NULL` (sequential).}

\item{pb}{Logical. Whether to display a progress bar. Default is `FALSE`.}

\item{...}{Additional arguments passed to `.f`.}
}
\value{
A list of results obtained by applying `.f` to each tuple from `.l`.
}
\description{
Applies a function `.f` over multiple aligned lists in `.l`.
Each element of `.l` should be a list or vector of the same length. Each call to `.f` receives one element from each list.
Supports parallel execution and progress display.
}
\examples{
# Fit a linear model for each response variable using the same predictor
df <- data.frame(
  y1 = rnorm(100),
  y2 = rnorm(100),
  x = rnorm(100)
)

# List of formulas and data
formulas <- list(y1 ~ x, y2 ~ x)
data_list <- list(df, df)

fmapn(list(formula = formulas, data = data_list),
      function(formula, data) lm(formula, data = data))

# Extract model summaries in parallel
models <- fmapn(
  list(formula = formulas, data = data_list),
  function(formula, data) summary(lm(formula, data = data))$r.squared
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fmapc.R
\name{fmapc}
\alias{fmapc}
\title{Apply a function column-wise with name access and parallelism}
\usage{
fmapc(.df, .f, ncores = NULL, pb = FALSE, ...)
}
\arguments{
\item{.df}{A data frame whose columns will be iterated over.}

\item{.f}{A function that takes two arguments: the column vector and its name.}

\item{ncores}{Integer. Number of cores to use for parallel processing. Default is `NULL` (sequential).}

\item{pb}{Logical. Whether to display a progress bar. Default is `FALSE`.}

\item{...}{Additional arguments passed to `.f`.}
}
\value{
A list of results obtained by applying `.f` to each column of `.df`.
}
\description{
Applies a function `.f` to each column of a data frame `.df`. Each call receives both
the column vector and its name, enabling name-aware column processing. Supports parallel execution and progress display.
}
\examples{
df <- data.frame(a = 1:3, b = 4:6)

# Apply a function that returns column mean and name
fmapc(df, function(x, name) {
  list(mean = mean(x), var = var(x), name = name)
})

# With progress and parallel execution
\donttest{
fmapc(df, function(x, name) mean(x), ncores = 2, pb = TRUE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package.R
\docType{package}
\name{functionals-package}
\alias{functionals}
\alias{functionals-package}
\title{functionals: Functional programming with parallelism and progress track in R}
\description{
This package provides a consistent and intuitive API for applying functions
over lists, vectors, and data frames, with built-in support for parallelism
and progress bars.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://CRAN.R-project.org/package=functionals}
  \item Report bugs at \url{https://github.com/ielbadisy/functionals/issues}
}

}
\author{
\strong{Maintainer}: Imad EL BADISY \email{elbadisyimad@gmail.com}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fapply.R
\name{fapply}
\alias{fapply}
\title{Apply a function over a list or vector with optional parallelism and progress}
\usage{
fapply(.x, .f, ncores = 1, pb = FALSE, cl = NULL, load_balancing = TRUE, ...)
}
\arguments{
\item{.x}{A list or atomic vector.}

\item{.f}{Function to apply.}

\item{ncores}{Number of cores to use (default: 1 = sequential).}

\item{pb}{Show progress bar? (default: FALSE). When enabled, progress is driven by
completed tasks. In parallel mode, updates occur as workers return results.}

\item{cl}{A cluster object (from parallel::makeCluster), or integer for core count.}

\item{load_balancing}{Logical. Use `parLapplyLB` if `TRUE` (default: `FALSE`).}

\item{...}{Additional arguments passed to `.f`.}
}
\value{
A list of results.
}
\description{
A lightweight and fast version of `lapply()` with support for multicore (Unix) and snow-style
clusters via `parallel`, with internal progress bar tracking and message suppression.
When `pb = TRUE`, progress advances when individual tasks complete, including in the
parallel paths, so counts reflect completed jobs rather than internal chunk boundaries.
Progress rendering is throttled for large workloads to keep console redraw overhead bounded,
while status output reports elapsed time and estimated time remaining.
}
\examples{
# Basic usage (sequential)
fapply(1:5, sqrt)

# With progress bar (sequential)
fapply(1:5, function(x) { Sys.sleep(0.1); x^2 }, pb = TRUE)

# Multicore on Unix (if available)
\donttest{
if (.Platform$OS.type != "windows") {
  fapply(1:10, sqrt, ncores = 2)
}
}

# With user-created cluster (portable across platforms)
\donttest{
cl <- parallel::makeCluster(2)
fapply(1:10, sqrt, cl = cl)
parallel::stopCluster(cl)
}

# Heavy computation example with exact completion-driven progress
\donttest{
heavy_fn <- function(x) { Sys.sleep(0.05); x^2 }
fapply(1:20, heavy_fn, ncores = 2, pb = TRUE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fmap.R
\name{fmap}
\alias{fmap}
\title{Functional mapping with optional parallelism and progress bars}
\usage{
fmap(.x, .f, ncores = NULL, pb = FALSE, ...)
}
\arguments{
\item{.x}{A list or atomic vector of elements to iterate over.}

\item{.f}{A function to apply to each element of `.x`. Can be a function or a string naming a function.}

\item{ncores}{Integer. Number of CPU cores to use for parallel processing. Default is `NULL` (sequential).}

\item{pb}{Logical. Whether to show a progress bar. Default is `FALSE`. In parallel
mode, progress updates as each completed task returns.}

\item{...}{Additional arguments passed to `.f`.}
}
\value{
A list of results, one for each element of `.x`.
}
\description{
Applies a function `.f` to each element of `.x`, with optional parallel processing and progress bar support.
Progress behavior is inherited from `fapply()`, including completion-driven updates
in the parallel execution paths, bounded redraw frequency for large workloads,
and elapsed/ETA status reporting.
}
\examples{
slow_fn <- function(x) { Sys.sleep(0.01); x^2 }
x <- 1:100

# Basic usage
fmap(x, slow_fn)

# With progress bar
fmap(x, slow_fn, pb = TRUE)

# With parallel execution (non-Windows)
\donttest{
if (.Platform$OS.type != "windows") {
  fmap(x, slow_fn, ncores = 2, pb = TRUE)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/frepeat.R
\name{frepeat}
\alias{frepeat}
\title{Repeat an expression or function call multiple times}
\usage{
frepeat(
  .x = NULL,
  times,
  expr,
  simplify = FALSE,
  ncores = NULL,
  pb = FALSE,
  ...
)
}
\arguments{
\item{.x}{Optional input passed to `expr` if `expr` is a function. Default is `NULL`.}

\item{times}{Integer. Number of repetitions.}

\item{expr}{A function or an unevaluated expression to repeat. If a function, it will be called `times` times.}

\item{simplify}{Logical. If `TRUE`, attempts to simplify the result using `simplify2array()`. Default is `FALSE`.}

\item{ncores}{Integer. Number of cores to use for parallel execution. Default is `NULL` (sequential).}

\item{pb}{Logical. Whether to display a progress bar. Default is `FALSE`.}

\item{...}{Additional arguments passed to the function `expr` if it is callable.}
}
\value{
A list of outputs (or a simplified array if `simplify = TRUE`) from evaluating `expr` multiple times.
}
\description{
Repeats an expression or function evaluation `times` times. If `expr` is a function, it is invoked
with optional input `.x` and additional arguments. If `expr` is a quoted expression, it is evaluated
in the parent environment. Supports parallel processing and optional simplification of results.
}
\note{
If `expr` is passed as a function call (not a function or quoted expression),
it will be evaluated immediately, not repeated. Use \code{function(...) \\{ ... \\}} or \code{quote(...)} instead.
}
\examples{
# Repeat a pure function call
frepeat(times = 3, expr = function() rnorm(1))

# Repeat a function with input `.x`
frepeat(.x = 10, times = 3, expr = function(x) rnorm(1, mean = x))

# Repeat an unevaluated expression (evaluated with `eval()`)
frepeat(times = 2, expr = quote(rnorm(1)))

# Simplify the output to an array
frepeat(times = 3, expr = function() rnorm(1), simplify = TRUE)

# Monte Carlo simulation: estimate coverage of a 95\% CI for sample mean
mc_result <- frepeat(
  times = 1000, simplify = TRUE, pb = TRUE, ncores = 1,
  expr = function() {
    sample <- rnorm(30, mean = 0, sd = 1)
    ci <- t.test(sample)$conf.int
    mean(ci[1] <= 0 & 0 <= ci[2])  # true mean inside the interval?
  }
)
mean(mc_result)  # estimated coverage

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcompose.R
\name{fcompose}
\alias{fcompose}
\title{Compose multiple functions}
\usage{
fcompose(...)
}
\arguments{
\item{...}{Functions to compose. Each must take a single argument and return an object
compatible with the next function in the chain.}
}
\value{
A new function equivalent to nested application of the input functions.
}
\description{
Create a new function by composing several functions, applied from right to left.
Equivalent to `f1(f2(f3(...)))`.
}
\examples{
square <- function(x) x^2
add1 <- function(x) x + 1

f <- fcompose(sqrt, square, add1)  # => sqrt(square(x + 1))
f(4)  # => sqrt((4 + 1)^2) = sqrt(25) = 5

# More compact
fcompose(log, exp)(2)  # log(exp(2)) = 2

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fmapr.R
\name{fmapr}
\alias{fmapr}
\title{Apply a function row-wise on a data frame with parallelism}
\usage{
fmapr(.df, .f, ncores = NULL, pb = FALSE, ...)
}
\arguments{
\item{.df}{A data frame whose rows will be iterated over.}

\item{.f}{A function applied to each row, which receives a named list.}

\item{ncores}{Integer. Number of cores to use for parallel processing. Default is `NULL` (sequential).}

\item{pb}{Logical. Whether to display a progress bar. Default is `FALSE`.}

\item{...}{Additional arguments passed to `.f`.}
}
\value{
A list of results returned by applying `.f` to each row as a list.
}
\description{
Applies a function `.f` to each row of a data frame `.df`, with optional parallelism and progress bar.
Each row is converted to a named list before being passed to `.f`, enabling flexible access to variables by name.
}
\examples{
df <- data.frame(name = c("Mister", "Hipster"), age = c(30, 25))

# Create personalized messages
fmapr(df, function(row) paste(row$name, "is", row$age, "years old"))

# Row-wise model formulas
formulas <- data.frame(
  response = c("y1", "y2"),
  predictor = c("x1", "x2"),
  stringsAsFactors = FALSE
)

fmapr(formulas, function(row) {
  reformulate(row$predictor, row$response)
})

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/floop.R
\name{floop}
\alias{floop}
\title{Functional loop with optional parallelism and progress bar}
\usage{
floop(.x, .f, ncores = 1, pb = FALSE, .capture = TRUE, ...)
}
\arguments{
\item{.x}{A vector or list of elements to iterate over.}

\item{.f}{A function to apply to each element of `.x`.}

\item{ncores}{Integer. Number of cores to use. Default is 1 (sequential).}

\item{pb}{Logical. Show a progress bar? Default is `FALSE`.}

\item{.capture}{Logical. Should results of `.f` be captured and returned? If `FALSE`, acts like a side-effect loop.}

\item{...}{Additional arguments passed to `.f`.}
}
\value{
A list of results if `.capture = TRUE`, otherwise returns `.x` invisibly.
}
\description{
`floop()` applies a function `.f` to each element of `.x`, optionally in parallel, and with an optional progress bar.
Unlike `fwalk()`, it can return results or be used purely for side effects (like a for-loop).
}
\examples{
# Functional loop that collects output
floop(1:3, function(i) i^2)

# Side-effect only loop (like for-loop with cat)
\donttest{
floop(1:5, function(i) cat(" Processing", i, "\n"),
      pb = TRUE, .capture = FALSE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/freduce.R
\name{freduce}
\alias{freduce}
\title{Functional reduce}
\usage{
freduce(
  .x,
  .f,
  .init = NULL,
  .right = FALSE,
  .accumulate = FALSE,
  .simplify = TRUE
)
}
\arguments{
\item{.x}{A vector or list to reduce.}

\item{.f}{A binary function to apply. Can be given as a function or quoted (e.g., `\`+\``).}

\item{.init}{Optional initial value passed to [Reduce()]. If `NULL`, reduction starts from the first two elements.}

\item{.right}{Logical. If `TRUE`, reduction is performed from right to left.}

\item{.accumulate}{Logical. If `TRUE`, returns a list of intermediate results (like a scan).}

\item{.simplify}{Logical. If `TRUE` and all intermediate results are length 1, the output is simplified to a vector.}
}
\value{
A single value (default) or a list/vector of intermediate results if `.accumulate = TRUE`.
}
\description{
Apply a binary function iteratively over a list or vector, reducing it to a single value or a sequence of intermediate results.
This is a wrapper around [Reduce()] that supports optional initial values, right-to-left evaluation,
accumulation of intermediate steps, and output simplification.
}
\examples{
freduce(1:5, `+`)                          # => 15
freduce(letters[1:4], paste0)             # => "abcd"
freduce(list(1, 2, 3), `*`)               # => 6
freduce(1:3, `+`, .init = 10)             # => 16
freduce(1:3, paste0, .right = TRUE)       # => "321"
freduce(1:4, `+`, .accumulate = TRUE)     # => c(1, 3, 6, 10)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcv.R
\name{fcv}
\alias{fcv}
\title{Functional Cross-Validation mapping}
\usage{
fcv(.splits, .f, ncores = NULL, pb = FALSE, ...)
}
\arguments{
\item{.splits}{A list of resample splits (e.g., from `rsample::vfold_cv()`).}

\item{.f}{A function to apply to each split. Typically expects a single `split` object.}

\item{ncores}{Integer. Number of cores to use for parallel processing. Default is `NULL` (sequential).}

\item{pb}{Logical. Whether to display a progress bar. Default is `FALSE`.}

\item{...}{Additional arguments passed to `.f`.}
}
\value{
A list of results returned by applying `.f` to each element of `.splits`.
}
\description{
Applies a user-defined function `.f` to each element of `.splits`, typically from
cross-validation objects such as `rsample::vfold_cv()`.
}
\examples{
if (requireNamespace("rsample", quietly = TRUE)) {
  set.seed(123)
  cv_splits <- rsample::vfold_cv(mtcars, v = 5)

  # Apply summary over training sets
  fcv(cv_splits$splits, function(split) {
    summary(rsample::analysis(split))
  })

  # With progress and parallel execution
  \donttest{
    fcv(cv_splits$splits, function(split) {
      summary(rsample::analysis(split))
    }, ncores = 2, pb = TRUE)
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fwalk.R
\name{fwalk}
\alias{fwalk}
\title{Walk over a vector or list with side effects}
\usage{
fwalk(.x, .f, ncores = NULL, pb = FALSE, ...)
}
\arguments{
\item{.x}{A list or atomic vector of elements to iterate over.}

\item{.f}{A function to apply to each element of `.x`. Should be called primarily for side effects.}

\item{ncores}{Integer. Number of cores to use for parallel processing. Default is `NULL` (sequential).}

\item{pb}{Logical. Whether to show a progress bar. Default is `FALSE`.}

\item{...}{Additional arguments passed to `.f`.}
}
\value{
Invisibly returns `.x`, like `purrr::walk()`.
}
\description{
Applies a function `.f` to each element of `.x`, typically for its side effects (e.g., printing, writing files).
This function is the side-effect-friendly equivalent of `fmap()`. Supports parallel execution and progress bar display.
}
\examples{
# Print each element
fwalk(1:3, print)

# Simulate writing files in parallel
\donttest{
fwalk(1:3, function(i) {
  cat(paste("Processing item", i, "\n"))
  Sys.sleep(0.5)
}, ncores = 2, pb = TRUE)
}

}
